% -------------------------------------------------------------------------
% 4K160 - Modeling and control of hybrid dynamical systems
% Example 1: Finding a quadratic Lyapunov function for a linear system
%            V(x)=x'Px;
% -------------------------------------------------------------------------
close all
clear all

% system dynamics
A = [-1 2 0;-3 -4 1;0 0 -2];

% LMI variable sdpvar defines the dimension of the unknown variable in the
% matrix inequality n lines times m columns. Square matrices are by default
% defined as symmetric.

Z = sdpvar(3,3);

% Having Z, which stands for the weight of the Lyapunov function, we are ready to define the constraints. 
% We do not use P here because the function "value" needs to be used later
% to convert Z into normal Matlab matrix. Rather than using P = value(P)
% which may lead to wrong re-assignments in Matlab, we shall use Z as
% variable and then P = value(Z) to retrieve the solution in matrix
% format.

Q = 0.1*eye(3); % Matrix related to the decrease rate

% Define the constraints
L = [A'*Z+Z*A + Q <= 0, Z>=1e-9]; % key command for imposing linear matrix conditions
% We use Z>= 1e-9 instead of Z>0, as SeDuMi cannot work with strict inequalities

% Solving the LMI conditions with SeDuMi: optimize creates the correct
% arguments and calls the SeDuMi solver to obtain the solution.

diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end
pause;

% value is just a function that converts the solution Z into Matlab standard matrix format.

P = value(Z)
eig(P)
disp('Eigenvalues are positive so P is ok, the system is stable')


pause;
%% How to minimize?
disp('Optimizing over elements of P')
pause;

% Define unknown variables
gamma = sdpvar(1,1); % defines a scalar variable
Z = sdpvar(3);

L = [A'*Z+Z*A+ Q <= 0,
     Z >= 1e-9,
     gamma*eye(3) >= Z, % No eigenvalues of Z larger than gamma, next minimze gamma
     %-gamma + 1 >= 0, gamma - 0.01 >= 0 % gamma between 0.1 and 1
     ];   
    

% besides L there is a second argument, which is not an inequality but a
% variable. This automatically solves the inequalities in L while optimizing
% the specified variable.
diagnostics = optimize(L,gamma);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end

% solutions are always retrieved using the function value.
P = value(Z)
disp('eigenvalues of P:'), eig(P)
opt = value(gamma)
