% -------------------------------------------------------------------------
% 4K160 - Modeling and control of hybrid dynamical systems
% Example 2: Checking stability of a PWL system (no S-procedure used)
% -------------------------------------------------------------------------
close all
clear all
clc 

yalmip clear

% System dynamics:
A1 = [-0.1 1; -5 -0.1];
A2 = [-0.1 5; -1 -0.1];

% LMI variables:
Z=sdpvar(2,2);

%% Search for a common quadratic Lyapunov function
% Lyapunov condition:
L1 = [A1'*Z+Z*A1<=-1e-9];
L2 = [A2'*Z+Z*A2<=-1e-9];

Lp = [Z>=1e-9];

% combine constraints into one object
L = [L1,L2,Lp]; 

% solve the LMI using SeDuMi:
diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end
 
pause;

% retrieve the solution:

P=value(Z)
eig(P)
disp('The LMI is not feasible  ... - no common quadratic Lyapunov function exists')
disp('Do not trust only on outcome "solver info", check always the eigenvalues!!')
disp('Use checkset(L) to look at the primal residuals: positive = feasible; some negative = not feasible')

pause;

checkset(L)

pause;

% check for stability system 1
optimize([L1,Lp]);
P=value(Z)
eig(P)
disp('P is ok, system 1 stable');

pause;

% check for stability system 2
optimize(L2+Lp);
P=value(Z)
eig(P)
disp('P is ok, system 2 stable');

pause;
clear all;

%% Search for a Piecewise quadratic one: P1, P2 (no S-procedure used!)
% First define the matrices Ei*x>=0 that define the regions where dynamics
%   A1, A2 are active

% Mode A1 is active in quadrant 1 and 3 defined by E1x>=0 and E3x>=0
A1 = [-0.1 1; -5 -0.1];
A2 = [-0.1 5; -1 -0.1];

E1=[-1 1; -1 -1]; E3=-E1;

% Mode A2 is active in quadrant 2 and 3 defined by E2x>=0 and E4x>=0

E2=[-1 1; 1 1]; E4=-E2;

% To ensure continuity we define Pi=Fi'*T*Fi where T is a 4x4 common matrix
%   and Fi is constructed for each quadrant i using the matrix Ei as follows (i=1,2,3,4):

F1 = [eye(2) ;E1]; 
F2 = [eye(2) ;E2];
F3 = [eye(2) ;E3]; 
F4 = [eye(2) ;E4];

% LMI variables: Z now stands for T

Z=sdpvar(4,4);

% Lyapunov condition:
Z1=F1'*Z*F1;
Z2=F2'*Z*F2;
Z3=F3'*Z*F3;
Z4=F4'*Z*F4;

% We only need to check two cases ... why? 
L1=[A1'*Z1+Z1*A1<=-1e-9];
L2=[A2'*Z2+Z2*A2<=-1e-9];
L3=[A1'*Z3+Z3*A1<=-1e-9];
L4=[A2'*Z4+Z4*A2<=-1e-9];

Lp1=[Z1>=1e-9];
Lp2=[Z2>=1e-9];
Lp3=[Z3>=1e-9];
Lp4=[Z4>=1e-9];

% Gathering all conditions in one: using the simple +
L=L1+L2+L3+L4+Lp1+Lp2+Lp3+Lp4;

% Solve the LMI using SeDuMi:

diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end

% retrieve the solutions:

T=value(Z)
P1=F1'*T*F1
eig(P1)
P2=F2'*T*F2
eig(P2)
P3=F3'*T*F3
eig(P3)
P4=F4'*T*F4
eig(P4)
disp('The PWL system is asymptotically stable')


