% -------------------------------------------------------------------------
% 4K160 - Modeling and control of hybrid dynamical systems
% Example 3: why is the S-procedure useful
% -------------------------------------------------------------------------
close all
clear all 
clc

yalmip clear

% System dynamics:
A1 = [ -25.1745   -8.0890;  17.3378    6.0470];
A2 = [-180.0290  147.4696;  522.8343 -503.3861];

%% First: common Lyapunov function
% LMI variables:
Z=sdpvar(2,2);

% Lyapunov condition:
L1=[A1'*Z+Z*A1<=-1e-9];
L2=[A2'*Z+Z*A2<=-1e-9];

Lp=[Z>=1e-9];

L=L1+L2+Lp;

% solve the LMI using SeDuMi:
diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end

% retrieve the solution:

P=value(Z)
disp('The LMI is not feasible ... - no common quadratic Lyapunov function exists')

pause;

clear all;

%% Search for a Piecewise quadratic one: P1, P2
% First define the matrices Ei*x>=0 that define the regions
% A1 is active in x1*x2<=0 
A1 = [ -25.1745   -8.0890;  17.3378    6.0470];
A2 = [-180.0290  147.4696;  522.8343 -503.3861];

% regions
E1=[-1 1; -1 -1];
E3=-E1;
E2=[-1 1; 1 1];
E4=-E2;

% To ensure continuity define Pi=Fi'*T*Fi
F1=[E1' eye(2)]';
F2=[E2' eye(2)]';
F3=[E3' eye(2)]';
F4=[E4' eye(2)]';

% LMI variables:
Z=sdpvar(4,4);

% Lyapunov condition:
Z1=F1'*Z*F1;
Z2=F2'*Z*F2;
Z3=F3'*Z*F3;
Z4=F4'*Z*F4;

L1=[A1'*Z1+Z1*A1<=-1e-9];
L2=[A2'*Z2+Z2*A2<=-1e-9];
L3=[A1'*Z3+Z3*A1<=-1e-9];
L4=[A2'*Z4+Z4*A2<=-1e-9];

Lp1=[Z1>=1e-9];
Lp2=[Z2>=1e-9];
Lp3=[Z3>=1e-9];
Lp4=[Z4>=1e-9];

L=L1+L2+L3+L4+Lp1+Lp2+Lp3+Lp4;

% solve the LMI using SeDuMi:
diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end

% retrieve the solutions:

T=value(Z)
P1=F1'*T*F1
eig(P1)
P2=F2'*T*F2
eig(P2)
P3=F3'*T*F3
eig(P3)
P4=F4'*T*F4
eig(P4)

disp('a PWQ function not found!')

pause;
clear all;

%% Search for a Piecewise quadratic one with S-procedure: P1, P2, U1, U2, W1, W2
% First define the matrices Ei*x>=0 that define the regions
A1 = [ -25.1745   -8.0890;  17.3378    6.0470];
A2 = [-180.0290  147.4696;  522.8343 -503.3861];

% regions
E1=[-1 1; -1 -1];
E3=-E1;
E2=[-1 1; 1 1];
E4=-E2;

% To ensure continuity define Pi=Fi'*T*Fi
F1=[E1' eye(2)]';
F2=[E2' eye(2)]';
F3=[E3' eye(2)]';
F4=[E4' eye(2)]';

% LMI variables:
Z=sdpvar(4,4);
U1=sdpvar(2,2);
U2=sdpvar(2,2);
U3=sdpvar(2,2);
U4=sdpvar(2,2);

W1=sdpvar(2,2);
W2=sdpvar(2,2);
W3=sdpvar(2,2);
W4=sdpvar(2,2);

% Lyapunov condition:
Z1=F1'*Z*F1;
Z2=F2'*Z*F2;
Z3=F3'*Z*F3;
Z4=F4'*Z*F4;

L1=[A1'*Z1+Z1*A1+E1'*U1*E1<=-1e-9];
L2=[A2'*Z2+Z2*A2+E2'*U2*E2<=-1e-9];
L3=[A1'*Z3+Z3*A1+E3'*U3*E3<=-1e-9];
L4=[A2'*Z4+Z4*A2+E4'*U4*E4<=-1e-9];

Lp1=[Z1-E1'*W1*E1>=1e-9];
Lp2=[Z2-E2'*W2*E2>=1e-9];
Lp3=[Z3-E3'*W3*E3>=1e-9];
Lp4=[Z4-E4'*W4*E4>=1e-9];

pos=[U1(:)>=0, U2(:)>=0, U3(:)>=0, U4(:)>=0, W1(:)>=0, W2(:)>=0, W3(:)>=0, W4(:)>=0];

L=L1+L2+L3+L4+Lp1+Lp2+Lp3+Lp4+pos;


%solve the LMI using SeDuMi:
diagnostics = optimize(L);
if diagnostics.problem == 0
 disp('Feasible')
elseif diagnostics.problem == 1
 disp('Infeasible')
else
 disp('Something else happened')
end

%retrieve the solutions:

T=value(Z)
P1=F1'*T*F1
eig(P1)
U1=value(U1)
W1=value(W1)
pause;

P2=F2'*T*F2
eig(P2)
U2=value(U2)
W2=value(W2)
pause;

P3=F3'*T*F3
eig(P3)
U3=value(U3)
W3=value(W3)
pause;

P4=F4'*T*F4
eig(P4)
U4=value(U4)
W4=value(W4)

disp('a PWQ function with S-procedure found!') 

